'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _arrayFlatten = require('array-flatten');

var _arrayFlatten2 = _interopRequireDefault(_arrayFlatten);

var _Logger = require('../Logger');

var _Logger2 = _interopRequireDefault(_Logger);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const log = _Logger2.default.child({
  namespace: 'normalizeAnonymousValuePlaceholders'
});

const anonymousePlaceholdersRegex = /\?/g;

/**
 * @see https://github.com/mysqljs/sqlstring/blob/f946198800a8d7f198fcf98d8bb80620595d01ec/lib/SqlString.js#L73
 */
// eslint-disable-next-line complexity

exports.default = (sql, values = []) => {
  if (!anonymousePlaceholdersRegex.test(sql)) {
    return {
      sql,
      values
    };
  }

  anonymousePlaceholdersRegex.lastIndex = 0;

  let chunkIndex = 0;
  let result = '';
  let match;
  let valueIndex = 0;
  let placeholderIndex = 0;

  // eslint-disable-next-line no-cond-assign
  while (match = anonymousePlaceholdersRegex.exec(sql)) {
    if (!values.hasOwnProperty(valueIndex)) {
      throw new Error('Value placeholder is missing a value.');
    }

    const value = values[valueIndex];

    valueIndex++;

    result += sql.slice(chunkIndex, match.index);
    chunkIndex = anonymousePlaceholdersRegex.lastIndex;

    // SELECT ?, [[[1,1],[1,1]]]; SELECT ($1, $2), ($3, $4)
    if (Array.isArray(value) && Array.isArray(value[0])) {
      const valueSets = [];

      let valueSetListSize = value.length;

      while (valueSetListSize--) {
        const placeholders = [];

        let setSize = value[0].length;

        while (setSize--) {
          placeholders.push('$' + ++placeholderIndex);
        }

        valueSets.push('(' + placeholders.join(', ') + ')');
      }

      result += valueSets.join(', ');

      // SELECT ?, [[1,1]]; SELECT ($1, $2)
    } else if (Array.isArray(value)) {
      const placeholders = [];

      let setSize = value.length;

      while (setSize--) {
        placeholders.push('$' + ++placeholderIndex);
      }

      result += '(' + placeholders.join(', ') + ')';
    } else {
      placeholderIndex++;

      result += '$' + placeholderIndex;
    }
  }

  if (values.length !== valueIndex) {
    throw new Error('Value number is greater than the placeholder count.');
  }

  if (chunkIndex === 0) {
    result = sql;
  } else if (chunkIndex < sql.length) {
    result += sql.slice(chunkIndex);
  }

  log.trace({
    sql: result,
    values
  }, 'normalized SQL');

  return {
    sql: result,
    values: (0, _arrayFlatten2.default)(values)
  };
};
//# sourceMappingURL=normalizeAnonymousValuePlaceholders.js.map